# -*- coding: utf-8 -*-
from channels_api.bindings import ResourceBindingBase, ResourceBinding

from .mixins import DjangoGraphqlBindingMixin, UnsubscribeMixin


class SubscriptionResourceBinding(DjangoGraphqlBindingMixin, ResourceBindingBase):
    # mark as abstract
    model = None


class ExtraResourceBinding(UnsubscribeMixin, ResourceBinding):
    # mark as abstract
    model = None



# -*- coding: utf-8 -*-
import copy
import inspect
import json

import rx
from channels import Group
from graphene import Field, Argument, Enum, String, ObjectType, Boolean, List, ID
from graphene.types.base import BaseOptions
from graphene.utils.str_converters import to_snake_case
from rest_framework.serializers import Serializer
from six import string_types

from .bindings import SubscriptionResourceBinding


class ActionSubscriptionEnum(Enum):
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    ALL_ACTIONS = "all_actions"


class OperationSubscriptionEnum(Enum):
    SUBSCRIBE = "subscribe"
    UNSUBSCRIBE = "unsubscribe"


class SubscriptionOptions(BaseOptions):
    output = None
    fields = None
    arguments = None
    model = None
    stream = None
    serializer_class = None
    queryset = None
    interfaces = ()


class Subscription(ObjectType):
    """
        Subscription Type Definition
    """

    ok = Boolean(description="Boolean field that return subscription request result.")
    error = String(description="Subscribe or unsubscribe operation request error .")
    stream = String(description="Stream name.")
    operation = OperationSubscriptionEnum(description="Subscription operation.")
    action = ActionSubscriptionEnum(description="Subscription action.")

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        serializer_class=None,
        stream=None,
        queryset=None,
        description="",
        **options
    ):

        assert inspect.isclass(serializer_class) and issubclass(
            serializer_class, Serializer
        ), 'You need to pass a valid Serializer subclass in {}.Meta, received "serializer_class = {}"'.format(
            cls.__name__, serializer_class
        )

        assert isinstance(
            stream, string_types
        ), 'You need to pass a valid string stream name in {}.Meta, received "{}"'.format(
            cls.__name__, stream
        )

        if queryset:
            assert serializer_class.Meta.model == queryset.model, (
                "The queryset model must correspond with the serializer_class model passed on Meta class, received "
                '"{}", expected "{}"'.format(
                    queryset.model.__name__, serializer_class.Meta.model.__name__
                )
            )

        description = description or "Subscription Type for {} model".format(
            serializer_class.Meta.model.__name__
        )

        _meta = SubscriptionOptions(cls)

        _meta.output = cls
        _meta.fields = None
        _meta.model = serializer_class.Meta.model
        _meta.stream = stream
        _meta.serializer_class = serializer_class

        serializer_fields = [
            (to_snake_case(field.strip("_")).upper(), to_snake_case(field))
            for field in list(_meta.serializer_class().get_fields().keys())
        ]
        model_fields_enum = Enum(
            "{}Fields".format(_meta.model.__name__),
            serializer_fields,
            description="Desired {} fields in subscription's  notification.".format(
                _meta.model.__name__
            ),
        )

        arguments = {
            "channel_id": Argument(
                String,
                required=True,
                description="Websocket's channel connection identification",
            ),
            "action": Argument(
                ActionSubscriptionEnum,
                required=True,
                description="Subscribe or unsubscribe action : (create, update or delete)",
            ),
            "operation": Argument(
                OperationSubscriptionEnum, required=True, description="Operation to do"
            ),
            "id": Argument(
                ID,
                description="ID field value that has the object to which you wish to subscribe",
            ),
            "data": List(model_fields_enum, required=False),
        }

        _meta.arguments = arguments

        super(Subscription, cls).__init_subclass_with_meta__(
            _meta=_meta, description=description, **options
        )

    @classmethod
    def model_label(cls):
        return u"{}.{}".format(
            cls._meta.model._meta.app_label.lower(),
            cls._meta.model._meta.object_name.lower(),
        )

    @classmethod
    def _group_name(cls, action, id=None):
        """ Formatting helper for group names. """
        if id:
            # return '{}-{}-{}-{}'.format(cls.model_label(), cls._meta.stream.lower(), action, id)
            return "{}-{}-{}".format(cls.model_label(), action, id)
        else:
            # return '{}-{}-{}'.format(cls.model_label(), cls._meta.stream.lower(), action)
            return "{}-{}".format(cls.model_label(), action)

    @classmethod
    def subscription_resolver(cls, root, info, **kwargs):
        # Manage the subscribe or unsubscribe operations
        action = kwargs.get("action")
        operation = kwargs.get("operation")
        data = kwargs.get("data", None)
        obj_id = kwargs.get("id", None)

        response = {
            "stream": cls._meta.stream,
            "operation": operation,
            "action": action,
        }

        try:
            channel = copy.copy(info.context.reply_channel)
            channel.name = u"daphne.response.{}".format(kwargs.get("channel_id"))

            if action == "all_actions":
                for act in ("create", "update", "delete"):
                    group_name = cls._group_name(act, id=obj_id)

                    if operation == "subscribe":
                        Group(group_name).add(channel)
                    elif operation == "unsubscribe":
                        Group(group_name).discard(channel)
            else:
                group_name = cls._group_name(action, id=obj_id)

                if operation == "subscribe":
                    Group(group_name).add(channel)
                elif operation == "unsubscribe":
                    Group(group_name).discard(channel)

            if data is not None:
                setattr(cls._meta.serializer_class.Meta, "only_fields", data)

            response.update(dict(ok=True, error=None))
            channel.send({"text": json.dumps(response)})

        except Exception as e:
            response.update(dict(ok=False, error=e.__str__()))

        return rx.Observable.from_([cls(**response)])

    @classmethod
    def get_binding(cls):
        class ResourceBinding(SubscriptionResourceBinding):
            model = cls._meta.model
            stream = cls._meta.stream
            serializer_class = cls._meta.serializer_class
            queryset = cls._meta.queryset

        return ResourceBinding

    @classmethod
    def Field(cls, *args, **kwargs):
        kwargs.update(
            {
                "description": "Subscription for {} model".format(
                    cls._meta.model.__name__
                )
            }
        )
        return Field(
            cls._meta.output,
            args=cls._meta.arguments,
            resolver=cls.subscription_resolver,
            **kwargs
        )



# -*- coding: utf-8 -*-
from channels.generic.websockets import WebsocketDemultiplexer


class GraphqlAPIDemultiplexer(WebsocketDemultiplexer):
    def connect(self, message, **kwargs):
        import json

        """ Forward connection to all consumers."""
        resp = json.dumps(
            {
                "channel_id": self.message.reply_channel.name.split(".")[-1],
                "connect": "success",
            }
        )
        self.message.reply_channel.send({"accept": True, "text": resp})
        for stream, consumer in self.consumers.items():
            kwargs["multiplexer"] = self.multiplexer_class(
                stream, self.message.reply_channel
            )
            consumer(message, **kwargs)



# -*- coding: utf-8 -*-
import json

from channels import Group
from channels_api import detail_action
from rest_framework.exceptions import ValidationError


class UnsubscribeMixin(object):
    @detail_action()
    def unsubscribe(self, pk, data, **kwargs):
        if "action" not in data:
            raise ValidationError("action required")
        action = data["action"]
        group_name = self._group_name(action, id=pk)
        Group(group_name).discard(self.message.reply_channel)
        return {"action": action}, 200


class DjangoGraphqlBindingMixin(object):
    def deserialize(self, message):
        body = json.loads(message["text"])
        self.request_id = body.get("request_id")
        action = body.get("action")
        data = body.get("data", None)
        pk = data.get("id", None)
        return action, pk, data

    def serialize(self, instance, action):
        payload = {
            "action": action,
            "data": self.serialize_data(instance),
            "model": self.model_label,
        }
        return payload

    def serialize_data(self, instance):
        data = self.get_serializer(instance).data
        only_fields = hasattr(self.get_serializer_class().Meta, "only_fields")
        if only_fields:
            data = {
                k: v
                for k, v in data.items()
                if k in self.get_serializer_class().Meta.only_fields
            }
        return data



import ast
import os
import re

from setuptools import setup

_name = "graphene_django_subscriptions"
_version_re = re.compile(r"VERSION\s+=\s+(.*)")

with open("{}/__init__.py".format(_name), "rb") as f:
    version = ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    version = ".".join([str(v) for v in version])
    version = version.split(".final")[0] if "final" in version else version


def get_packages():
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(_name)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


setup(
    name=_name,
    version=version,
    description="Graphene-Django-Subscriptions add subscriptions support to graphene-django through "
    "Channels module",
    long_description=open("README.rst").read(),
    url="https://github.com/eamigo86/graphene-django-subscriptions",
    author="Ernesto Perez Amigo",
    author_email="eamigop86@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords="api graphql subscription rest graphene django channels",
    packages=get_packages(),
    install_requires=[
        "graphene-django>=2.0.0",
        "graphene-django-extras>=0.4.8",
        "channels-api>=0.4.1",
    ],
    include_package_data=True,
    zip_safe=False,
    platforms="any",
)



# -*- coding: utf-8 -*-
from graphene import get_version

from .consumers import GraphqlAPIDemultiplexer
from .middleware import SubscriptionMiddleware
from .subscription import Subscription

__author__ = "Ernesto"

VERSION = (0, 0, 9, "final", "")

__version__ = get_version(VERSION)

__all__ = (
    "__version__",
    "Subscription",
    "GraphqlAPIDemultiplexer",
    "SubscriptionMiddleware",
)



# -*- coding: utf-8 -*-
from promise import is_thenable


class SubscriptionMiddleware:
    def resolve(self, next, root, info, **kwargs):
        result = next(root, info, **kwargs)
        if info.operation.operation == "subscription" and is_thenable(result):
            return result.get()
        return result
